import abc
import logging
from dataclasses import dataclass, field
from datetime import datetime
from enum import Enum
from typing import Literal
from uuid import uuid4

from ..event import Event
from .detections import Category, Detections
from .image import Image


class Autoupload(Enum):
    """Configuration options for image auto-upload behavior to the Learning Loop"""

    FILTERED = 'filtered'
    """Upload mode for images with novel detections within uncertainty thresholds (default)"""

    DISABLED = 'disabled'
    """Upload mode where no images are auto-uploaded"""

    ALL = 'all'
    """Upload mode where every processed image is uploaded"""


class DetectorException(Exception):
    """An exception that is raised by a detector."""


@dataclass(slots=True, kw_only=True)
class DetectorInfo:
    """Information about the detector."""

    operation_mode: str
    """The operation mode of the detector node."""

    state: Literal['idle', 'online', 'detecting'] | None
    """The state of the detector node."""

    organization: str | None = None
    """The owner organization of the model."""

    project: str | None = None
    """The project of the model."""

    current_version: str | None = None
    """The currently used version of the model."""

    categories: list[Category] = field(default_factory=list)
    """The categories used in the model."""

    resolution: int | None = None
    """The resolution of the model (width and height of the image after preprocessing in pixels)."""

    target_version: str | None = None
    """The target version of the detector node."""

    version_control: Literal['follow_loop', 'specific_version', 'pause']
    """The version control mode of the detector node."""


@dataclass(slots=True, kw_only=True)
class ModelVersioningInfo:
    current_version: str
    """The version of the model currently used by the detector."""

    target_version: str
    """The target model version set in the detector."""

    loop_version: str
    """The target model version specified by the loop."""

    local_versions: list[str]
    """The locally available versions of the model."""

    version_control: str
    """The version control mode."""


class Detector(abc.ABC):
    """A detector allows detecting objects in images.

    It also holds an upload queue for sending images with uncertain results to an active learning infrastructure
    like the [Zauberzeug Learning Loop](https://zauberzeug.com/products/learning-loop).
    """

    def __init__(self, *, name: str | None = None) -> None:
        self.name = name or str(uuid4())

        self.NEW_DETECTIONS = Event()
        """detection on an image is completed (argument: image)"""

        self.log = logging.getLogger('rosys.detector')

    @abc.abstractmethod
    async def detect(self,
                     image: Image,
                     *,
                     autoupload: Autoupload = Autoupload.FILTERED,
                     tags: list[str] | None = None,
                     source: str | None = None,
                     creation_date: datetime | str | None = None,
                     ) -> Detections | None:
        """Runs detections on the image and fills the ``image.detections`` property.

        The parameters ``tags``, ``source``, and ``creation_date`` are added as metadata if the image is uploaded.

        Note that the hardware detector uses a lazy strategy to schedule the inference tasks.
        In particular a queue with a maximum size of 1 is used.
        This means if the detector is busy, the image is not processed immediately, but queued up.
        If the ``detect`` function is called again, the queued image is dropped and the new image is queued instead.
        In this case this method returns ``None``.

        :return: the detections found in the image.
        :raises DetectorException: if the detection fails.
        """

    @abc.abstractmethod
    async def upload(self,
                     image: Image,
                     *,
                     tags: list[str] | None = None,
                     source: str | None = None,
                     creation_date: datetime | str | None = None,
                     ) -> None:
        """Uploads the image to the Learning Loop.

        The parameters `tags`, `source`, and `creation_date` are added as metadata.
        If the image has detections, they are also uploaded.

        :raises DetectorException: if the upload fails.
        """

    @abc.abstractmethod
    async def fetch_detector_info(self) -> DetectorInfo:
        """Retrieve information about the detector.

        :return: information about the detector.
        :raises DetectorException: if the about information cannot be retrieved.
        """

    @abc.abstractmethod
    async def fetch_model_version_info(self) -> ModelVersioningInfo:
        """Retrieve information about the model version and versioning mode.

        :return: the information about the model versioning as data class.
        :raises DetectorException: if the detector is not connected or the information cannot be retrieved.
        """

    @abc.abstractmethod
    async def set_model_version(self, version: Literal['follow_loop', 'pause'] | str) -> None:
        """Set the model version or versioning mode.

        Set to "follow_loop" to automatically update the model version to the latest version in the learning loop.
        Set to "pause" to stop automatic updates and keep the current model version.
        Set to a version number (e.g. "1.2") to use a specific version.

        :raises DetectorException: if the version control mode is not valid or the version could not be set.
        """



from __future__ import annotations

from dataclasses import dataclass

import numpy as np
from nicegui import ui

from .. import rosys
from ..driving.driver import PoseProvider
from ..event import Event
from ..geometry import Rotation
from .module import Module, ModuleHardware, ModuleSimulation
from .robot_brain import RobotBrain


@dataclass(slots=True, kw_only=True)
class ImuMeasurement:
    """Imu measurement data with corrected and uncorrected angles and angular velocities in radians."""
    time: float
    gyro_calibration: float
    rotation: Rotation
    angular_velocity: Rotation


class Imu(Module):
    """A module that provides measurements from an IMU."""

    def __init__(self, offset_rotation: Rotation | None = None, **kwargs) -> None:
        super().__init__(**kwargs)
        self.offset_rotation = offset_rotation or Rotation.zero()
        self.last_measurement: ImuMeasurement | None = None

        self.NEW_MEASUREMENT = Event()
        """a new measurement has been received (argument: ImuMeasurement)"""

    def _emit_measurement(self, gyro_calibration: float, raw_rotation: Rotation, time: float) -> None:
        new_measurement = ImuMeasurement(
            time=time,
            gyro_calibration=gyro_calibration,
            rotation=raw_rotation * self.offset_rotation.T,
            angular_velocity=Rotation.zero(),
        )
        if self.last_measurement is not None:
            d_rotation = new_measurement.rotation * self.last_measurement.rotation.T
            d_roll, d_pitch, d_yaw = d_rotation.euler
            d_t = time - self.last_measurement.time
            new_measurement.angular_velocity = Rotation.from_euler(d_roll / d_t, d_pitch / d_t, d_yaw / d_t)
        self.last_measurement = new_measurement
        self.NEW_MEASUREMENT.emit(new_measurement)

    def developer_ui(self) -> None:
        ui.label('IMU').classes('text-center text-bold')
        with ui.row().classes('gap-0 w-56'):
            with ui.column().classes('gap-y-0 w-1/3'):
                ui.label('Roll:')
                ui.label().bind_text_from(self, 'last_measurement',
                                          lambda m: f'{np.rad2deg(m.rotation.roll):.2f}°' if m is not None else 'N/A')
                ui.label().bind_text_from(self, 'last_measurement',
                                          lambda m: f'{np.rad2deg(m.angular_velocity.roll):.2f}°/s' if m is not None and m.angular_velocity.roll is not None else 'N/A')
            with ui.column().classes('gap-y-0 w-1/3'):
                ui.label('Pitch:')
                ui.label().bind_text_from(self, 'last_measurement',
                                          lambda m: f'{np.rad2deg(m.rotation.pitch):.2f}°' if m is not None else 'N/A')
                ui.label().bind_text_from(self, 'last_measurement',
                                          lambda m: f'{np.rad2deg(m.angular_velocity.pitch):.2f}°/s' if m is not None and m.angular_velocity.pitch is not None else 'N/A')
            with ui.column().classes('gap-y-0 w-1/3'):
                ui.label('Yaw:')
                ui.label().bind_text_from(self, 'last_measurement',
                                          lambda m: f'{np.rad2deg(m.rotation.yaw):.2f}°' if m is not None else 'N/A')
                ui.label().bind_text_from(self, 'last_measurement',
                                          lambda m: f'{np.rad2deg(m.angular_velocity.yaw):.2f}°/s' if m is not None and m.angular_velocity.yaw is not None else 'N/A')
        ui.label().bind_text_from(self, 'last_measurement',
                                  lambda m: f'Gyro-Calibration: {m.gyro_calibration:.0f}' if m is not None else 'N/A')


class ImuHardware(Imu, ModuleHardware):
    """A hardware module that handles the communication with an IMU."""

    def __init__(self, robot_brain: RobotBrain, name: str = 'imu', **kwargs) -> None:
        self.name = name
        self.lizard_code = f'{name} = Imu()'
        self.core_message_fields = [
            f'{name}.cal_gyr',
            f'{name}.quat_w:4',
            f'{name}.quat_x:4',
            f'{name}.quat_y:4',
            f'{name}.quat_z:4',
        ]
        super().__init__(robot_brain=robot_brain, lizard_code=self.lizard_code, core_message_fields=self.core_message_fields, **kwargs)

    def handle_core_output(self, time: float, words: list[str]) -> None:
        gyro_calibration = float(words.pop(0))
        raw_rotation = Rotation.from_quaternion(
            float(words.pop(0)),
            float(words.pop(0)),
            float(words.pop(0)),
            float(words.pop(0)),
        )

        if gyro_calibration < 1.0:
            return
        self._emit_measurement(gyro_calibration, raw_rotation, time)


class ImuSimulation(Imu, ModuleSimulation):
    """Simulation of an IMU."""

    def __init__(self, *,
                 wheels: PoseProvider,
                 interval: float = 0.1,
                 roll_noise: float = 0.0,
                 pitch_noise: float = 0.0,
                 yaw_noise: float = 0.0,
                 **kwargs) -> None:
        super().__init__(**kwargs)
        self.wheels = wheels
        self._roll_noise = roll_noise
        self._pitch_noise = pitch_noise
        self._yaw_noise = yaw_noise
        rosys.on_repeat(self.simulate, interval)

    def simulate(self) -> None:
        roll = np.random.normal(0, self._roll_noise)
        pitch = np.random.normal(0, self._pitch_noise)
        yaw = np.random.normal(self.wheels.pose.yaw, self._yaw_noise)
        self._emit_measurement(3.0, Rotation.from_euler(roll, pitch, yaw), rosys.time())

    def developer_ui(self) -> None:
        super().developer_ui()
        with ui.column().classes('gap-y-0'):
            ui.number(label='Roll Noise', format='%.3f', prefix='± ', suffix='°') \
                .bind_value(self, '_roll_noise', forward=np.deg2rad, backward=np.rad2deg).classes('w-4/5')
            ui.number(label='Pitch Noise', format='%.3f', prefix='± ', suffix='°') \
                .bind_value(self, '_pitch_noise', forward=np.deg2rad, backward=np.rad2deg).classes('w-4/5')
            ui.number(label='Yaw Noise', format='%.3f', prefix='± ', suffix='°') \
                .bind_value(self, '_yaw_noise', forward=np.deg2rad, backward=np.rad2deg).classes('w-4/5')



from __future__ import annotations

import logging
import math
from typing import TYPE_CHECKING

import cv2
import numpy as np
from fastapi import Response
from nicegui import app

from .. import run
from .calibration import Calibration
from .image import Image

if TYPE_CHECKING:
    from .camera import Camera

log = logging.getLogger('rosys.image_route')


def create_image_route(camera: Camera) -> None:
    placeholder_url = '/' + camera.base_path + '/placeholder'
    timestamp_url = '/' + camera.base_path + '/{timestamp}'
    undistorted_url = '/' + camera.base_path + '/{timestamp}/undistorted'

    app.remove_route(placeholder_url)
    app.remove_route(timestamp_url)
    app.remove_route(undistorted_url)

    async def get_camera_image(timestamp: str,
                               shrink: float = 1.0,
                               max_dimension: float | None = None,
                               fast: bool = True,
                               compression: int = 60) -> Response:
        return await _get_image(camera, timestamp,
                                shrink=shrink,
                                max_dimension=max_dimension,
                                undistort=False,
                                fast=fast,
                                compression=compression)

    async def get_camera_image_undistorted(timestamp: str,
                                           shrink: float = 1.0,
                                           max_dimension: float | None = None,
                                           fast: bool = True,
                                           compression: int = 60) -> Response:
        return await _get_image(camera,
                                timestamp,
                                shrink=shrink,
                                max_dimension=max_dimension,
                                undistort=True,
                                fast=fast,
                                compression=compression)

    app.add_api_route(placeholder_url, _get_placeholder)
    app.add_api_route(timestamp_url, get_camera_image)
    app.add_api_route(undistorted_url, get_camera_image_undistorted)


async def _get_placeholder(shrink: int = 1) -> Response:
    return Response(content=Image.create_placeholder('no image', shrink=shrink).data, media_type='image/jpeg')


async def _get_image(camera: Camera,
                     timestamp: str, *,
                     shrink: float,
                     max_dimension: float | None,
                     undistort: bool,
                     fast: bool,
                     compression: int) -> Response:
    try:
        if undistort and getattr(camera, 'calibration', None) is None:
            return Response(content='Camera is not calibrated', status_code=404)
        jpeg = await _try_get_jpeg(camera,
                                   timestamp,
                                   shrink=shrink,
                                   max_dimension=max_dimension,
                                   undistort=undistort,
                                   fast=fast,
                                   compression=compression)
        if not jpeg:
            return Response(content='Image not found', status_code=404)
        return Response(content=jpeg, headers={'cache-control': 'max-age=7776000'}, media_type='image/jpeg')
    except Exception:
        log.exception('could not get image')
        raise


async def _try_get_jpeg(camera: Camera,
                        timestamp: str, *,
                        shrink: float,
                        max_dimension: float | None,
                        undistort: bool,
                        fast: bool,
                        compression: int) -> bytes | None:
    for image in reversed(camera.images):
        if str(image.time) == timestamp and image.data is not None:
            shrink_from_max = max(image.size.width, image.size.height) / max_dimension if max_dimension else shrink

            shrink = max(1, shrink, shrink_from_max)

            if shrink == 1 and not undistort and compression == 60:
                return image.data

            calibration = camera.calibration if undistort else None  # type: ignore
            return await run.cpu_bound(_process, image.data, calibration, shrink, undistort, fast, compression)

    return None


def _process(data: bytes,
             calibration: Calibration | None,
             shrink: float,
             undistort: bool,
             fast: bool,
             compression: int) -> bytes | None:
    array = np.frombuffer(data, dtype=np.uint8)
    if array is None:
        return None
    image_array = cv2.imdecode(array, cv2.IMREAD_COLOR)

    if undistort:
        assert calibration is not None
        image_array = calibration.undistort_array(image_array, crop=True)
        if image_array is None or image_array.size == 0:
            logging.warning('undistort_array returned an empty image')
            return None

    if shrink != 1.0:
        if fast:
            shrink_factor = math.ceil(shrink)
            image_array = image_array[::shrink_factor, ::shrink_factor]
        else:
            new_width = int(image_array.shape[1] / shrink)
            new_height = int(image_array.shape[0] / shrink)
            # INTER_AREA is optimal for downsampling
            image_array = cv2.resize(image_array, (new_width, new_height), interpolation=cv2.INTER_AREA)

    _, encoded_image = cv2.imencode('.jpg', image_array, [int(cv2.IMWRITE_JPEG_QUALITY), compression])
    return encoded_image.tobytes()



import json
from collections.abc import Callable
from pathlib import Path

from nicegui import app, events, ui
from starlette.responses import FileResponse

from ..helpers import invoke
from . import registry


def export_button(title: str = 'Export', route: str = '/export', tmp_filepath: Path = Path('/tmp/export.json')) -> ui.button:
    @app.get(route)
    def get_export() -> FileResponse:
        registry.write_export(tmp_filepath)
        return FileResponse(tmp_filepath, filename='export.json')
    return ui.button(title, on_click=lambda: ui.download(route[1:]))


def import_button(title: str = 'Import', after_import: Callable | None = None) -> ui.button:
    async def restore_from_file(e: events.UploadEventArguments) -> None:
        await registry.restore_from_export(json.load(e.content))
        dialog.close()
        if after_import is not None:
            await invoke(after_import)
    with ui.dialog() as dialog, ui.card():
        ui.upload(on_upload=restore_from_file)
    return ui.button(title, on_click=dialog.open)



import asyncio
from datetime import datetime
from typing import Any, Literal

import socketio
import socketio.exceptions

from .. import persistence, rosys
from ..helpers import LazyWorker
from .detections import BoxDetection, Category, Detections, PointDetection, SegmentationDetection
from .detector import Autoupload, Detector, DetectorException, DetectorInfo, ModelVersioningInfo
from .image import Image


class DetectorHardware(Detector):
    """This detector communicates with a [YOLO detector](https://hub.docker.com/r/zauberzeug/yolov5-detector) via Socket.IO.

    It automatically connects and reconnects, submits and receives detections and sends images
    that should be uploaded to the [Zauberzeug Learning Loop](https://zauberzeug.com/products/learning-loop).

    Note: Images must be smaller than ``MAX_IMAGE_SIZE`` bytes (default: 10 MB).
    """
    MAX_IMAGE_SIZE = 10 * 1024 * 1024

    def __init__(self,
                 *,
                 port: int = 8004,
                 name: str | None = None,
                 auto_disconnect: bool = True) -> None:
        super().__init__(name=name)

        websocket_options = {
            'max_msg_size': self.MAX_IMAGE_SIZE + 1000,
        }
        self.sio = socketio.AsyncClient(websocket_extra_options=websocket_options)
        self.lazy_worker = LazyWorker()
        self.port = port
        self.auto_disconnect = auto_disconnect
        self.timeout_count = 0

        @self.sio.on('disconnect')
        def on_sio_disconnect() -> None:
            self.log.warning('sio disconnect on port %s', port)

        @self.sio.on('connect_error')
        def on_sio_connect_error(err) -> None:
            self.log.warning('sio connect error on %s: %s', port, err)

        rosys.on_repeat(self._ensure_connection, 10.0)

    @property
    def is_connected(self) -> bool:
        return self.sio.connected

    async def _ensure_connection(self) -> None:
        if not self.is_connected:
            self.log.info('trying reconnect %s', self.name)
            if not await self.connect():
                self.log.error('connection to %s at port %s failed; trying again', self.name, self.port)

    async def connect(self) -> bool:
        try:
            url = f'ws://localhost:{self.port}'
            self.log.info('connecting to detector at %s', url)
            await self.sio.connect(url, socketio_path='/ws/socket.io', wait_timeout=3.0)
            self.log.info('connected successfully')
            return True
        except Exception:
            return False

    async def disconnect(self) -> None:
        await self.sio.disconnect()

    async def upload(self,
                     image: Image,
                     *,
                     tags: list[str] | None = None,
                     source: str | None = None,
                     creation_date: datetime | str | None = None
                     ) -> None:
        if not self.is_connected:
            self.log.error('Upload failed: detector is not connected')
            raise DetectorException('detector is not connected')

        if not image.data:
            self.log.error('Upload failed: image data is empty')
            raise DetectorException('image data is empty')

        if len(image.data) > self.MAX_IMAGE_SIZE:
            self.log.error('Upload failed: image too large: %s', len(image.data))
            raise DetectorException('image too large')

        tags = tags or []
        try:
            self.log.info('Upload detections to port %s', self.port)
            data_dict: dict[str, Any] = {'image': image.data, 'mac': image.camera_id}
            detections = image.get_detections(self.name)
            if detections is not None:
                detections_dict = detections.to_dict()
                detections_dict['box_detections'] = _box_detections_to_int(detections_dict.pop('boxes'))
                detections_dict['point_detections'] = detections_dict.pop('points')
                detections_dict['segmentation_detections'] = detections_dict.pop('segmentations')
                data_dict['detections'] = detections_dict
            data_dict['source'] = source
            data_dict['creation_date'] = _creation_date_to_isoformat(creation_date)
            await self.sio.emit('upload', data_dict)

        except Exception as e:
            self.log.error('Upload failed with exception: %s', e)
            raise DetectorException('Upload failed') from e

    async def detect(self,
                     image: Image,
                     *,
                     autoupload: Autoupload = Autoupload.FILTERED,
                     tags: list[str] | None = None,
                     source: str | None = None,
                     creation_date: datetime | str | None = None,
                     ) -> Detections | None:
        if not self.is_connected:
            self.log.error('Detection failed: detector is not connected')
            raise DetectorException('detector is not connected')

        if image.is_broken:
            self.log.error('Detection failed: image is broken')
            raise DetectorException('image is broken')

        assert len(image.data or []) < self.MAX_IMAGE_SIZE, f'image too large: {len(image.data or [])}'
        tags = tags or []
        try:
            detections = await self.lazy_worker.run(self._detect(image, autoupload, tags, source, creation_date))
            if detections is not None:
                image.set_detections(self.name, detections)
                self.NEW_DETECTIONS.emit(image)
            return detections
        except asyncio.exceptions.CancelledError:
            raise DetectorException('Detection cancelled') from None

    async def _detect(self,
                      image: Image,
                      autoupload: Autoupload,
                      tags: list[str],
                      source: str | None = None,
                      creation_date: datetime | str | None = None,
                      ) -> Detections:
        try:
            response = await self.sio.call('detect', {
                'image': image.data,
                'mac': image.camera_id,
                'autoupload': autoupload.value,
                'tags': tags,
                'source': source,
                'creation_date': _creation_date_to_isoformat(creation_date),
            }, timeout=3)
        except socketio.exceptions.TimeoutError:
            self.timeout_count += 1
            if self.timeout_count > 5 and self.auto_disconnect:
                self.log.error('Detection timed out 5 times in a row. Disconnecting from detector %s', self.name)
                await self.disconnect()
            raise DetectorException('Detection timeout') from None
        except Exception as e:
            raise DetectorException('Detection failed') from e

        if not isinstance(response, dict):
            raise DetectorException('Invalid response from detector')
        try:
            detections = Detections(
                boxes=[persistence.from_dict(BoxDetection, d) for d in response.get('box_detections', [])],
                points=[persistence.from_dict(PointDetection, d) for d in response.get('point_detections', [])],
                segmentations=[persistence.from_dict(SegmentationDetection, d)
                               for d in response.get('segmentation_detections', [])],
            )
        except Exception as e:
            raise DetectorException('Failed to parse detections') from e

        self.timeout_count = 0
        if image.is_broken:  # NOTE: image can be marked broken while detection is underway
            raise DetectorException('Image is broken')

        return detections

    def __str__(self) -> str:
        return f'{type(self).__name__} ({"connected" if self.is_connected else "disconnected"})'

    async def fetch_detector_info(self) -> DetectorInfo:
        if not self.is_connected:
            raise DetectorException('detector is not connected')

        try:
            response = await self.sio.call('about', timeout=5)
        except socketio.exceptions.TimeoutError:
            self.log.error('Communication timeout for detector %s', self.name)
            raise DetectorException('Communication timeout') from None
        except Exception as e:
            raise DetectorException('Communication failed') from e

        if not isinstance(response, dict):
            raise DetectorException('Invalid response from detector')

        try:
            model_info_dict = response.get('model_info', {})
            return DetectorInfo(
                operation_mode=response['operation_mode'],
                state=response.get('state'),
                organization=model_info_dict.get('organization'),
                project=model_info_dict.get('project'),
                current_version=model_info_dict.get('version'),
                categories=[
                    Category(uuid=c['id'],
                             name=c['name'],
                             color=c.get('color'),
                             category_type=c.get('type'))
                    for c in model_info_dict.get('categories', [])
                ],
                resolution=model_info_dict.get('resolution'),
                target_version=response.get('target_model'),
                version_control=response['version_control'],
            )
        except Exception as e:
            raise DetectorException('Failed to parse detector info') from e

    async def fetch_model_version_info(self) -> ModelVersioningInfo:
        try:
            response = await self.sio.call('get_model_version', timeout=5)
        except socketio.exceptions.TimeoutError:
            self.log.error('Communication timeout for detector %s', self.name)
            raise DetectorException('Communication timed out') from None
        except Exception as e:
            raise DetectorException('Communication failed') from e

        if not isinstance(response, dict):
            raise DetectorException('Invalid response from detector')

        try:
            return ModelVersioningInfo(
                current_version=response['current_version'],
                target_version=response['target_version'],
                loop_version=response['loop_version'],
                local_versions=response['local_versions'],
                version_control=response['version_control'],
            )
        except Exception as e:
            raise DetectorException('Failed to parse model versioning info') from e

    async def set_model_version(self, version: Literal['follow_loop', 'pause'] | str) -> None:
        if version not in ['follow_loop', 'pause'] and not version.replace('.', '').isdigit():
            raise DetectorException(f'invalid version control mode: {version} '
                                    f'(allowed: "follow_loop", "pause" or a version number like "1.2")')

        try:
            response = await self.sio.call('set_model_version_mode', version, timeout=5)
        except socketio.exceptions.TimeoutError:
            self.log.error('Communication timeout for detector %s', self.name)
            raise DetectorException('Communication timeout') from None
        except Exception as e:
            self.log.error('Communication failed for detector %s: %s', self.name, e)
            raise DetectorException('Communication failed') from e

        if not isinstance(response, dict):
            raise DetectorException('Failed to set model version mode')

        if response.get('status') != 'OK':
            error_message = response.get('error', 'unknown error')
            raise DetectorException(f'Failed to set model version mode: {error_message}')

    async def soft_reload(self) -> None:
        """Trigger a soft reload of the detector.

        :raises DetectorException: if the communication fails.
        """
        try:
            await self.sio.call('soft_reload', timeout=5)
        except socketio.exceptions.TimeoutError:
            self.log.error('Communication timeout for detector %s', self.name)
            raise DetectorException('Communication timeout') from None
        except Exception as e:
            self.log.error('Communication failed for detector %s: %s', self.name, e)
            raise DetectorException('Communication failed') from e


def _box_detections_to_int(detections: list[dict]) -> list[dict]:
    for detection in detections:
        detection['x'] = int(detection['x'])
        detection['y'] = int(detection['y'])
        detection['width'] = int(detection['width'])
        detection['height'] = int(detection['height'])
    return detections


def _creation_date_to_isoformat(creation_date: datetime | str | None) -> str | None:
    if isinstance(creation_date, str):
        try:
            datetime.fromisoformat(creation_date)  # validate format
        except ValueError as e:
            raise ValueError(f'creation_date string "{creation_date}" is not in valid ISO format') from e
        return creation_date
    elif isinstance(creation_date, datetime):
        return creation_date.isoformat()
    return None



from collections.abc import Awaitable, Callable

from .axis_mjpeg_device import AxisMjpegDevice
from .mjpeg_device import MjpegDevice
from .motec_mjpeg_device import MotecMjpegDevice
from .vendors import VendorType, mac_to_vendor


class MjpegDeviceFactory:
    @staticmethod
    def create(mac: str, ip: str, *,
               index: int | None = None,
               username: str | None = None,
               password: str | None = None,
               control_port: int | None = None,
               on_new_image_data: Callable[[bytes, float], Awaitable | None]) -> MjpegDevice:

        if mac_to_vendor(mac) == VendorType.AXIS:
            return AxisMjpegDevice(mac, ip,
                                   index=index, username=username,
                                   password=password, on_new_image_data=on_new_image_data)

        if mac_to_vendor(mac) == VendorType.MOTEC:
            return MotecMjpegDevice(mac, ip,
                                    username=username, password=password,
                                    control_port=control_port, on_new_image_data=on_new_image_data)

        raise ValueError(f'Unknown vendor for mac="{mac}"')



from __future__ import annotations

from dataclasses import asdict, dataclass, field
from typing import Literal

import numpy as np

from ..geometry import Point


@dataclass(slots=True, kw_only=True)
class Category:
    uuid: str
    name: str
    color: str | None = None
    category_type: Literal['box', 'point', 'segmentation', 'classification'] | None = None


@dataclass(slots=True, kw_only=True)
class Detection:
    category_name: str
    model_name: str
    confidence: float
    x: float
    y: float
    uuid: str = ''  # NOTE this is used to keep track of simulated objects

    @property
    def cx(self) -> float:
        raise NotImplementedError()

    @property
    def cy(self) -> float:
        raise NotImplementedError()

    @property
    def center(self) -> Point:
        return Point(x=self.cx, y=self.cy)


@dataclass(slots=True, kw_only=True)
class BoxDetection(Detection):
    width: float
    height: float

    @property
    def cx(self) -> float:
        return self.x + self.width / 2

    @property
    def cy(self) -> float:
        return self.y + self.height / 2

    def __str__(self) -> str:
        return f'x:{self.x:.0f} y:{self.y:.0f}, w:{self.width:.0f} h:{self.height:.0f} cat:{self.category_name} conf:{self.confidence:.2f}'

    def to_svg(self, shrink: int = 1) -> str:
        color = 'red'
        x = self.x / shrink
        y = self.y / shrink
        return (
            f'<rect x="{x}" y="{y}" width="{self.width/shrink}" height="{self.height/shrink}" stroke-width="2" '
            f'stroke="{color}" fill="none" />'
            f'<text x="{x}" y="{y - 7}" text-anchor="start" stroke="{color}" fill="{color}" font-size="10">'
            f'{self.category_name} ({self.confidence:.0%})</text>'
        )

    def intersection_over_union(self, other_detection: BoxDetection) -> float:
        # https://www.pyimagesearch.com/2016/11/07/intersection-over-union-iou-for-object-detection/
        xA = max(self.x, other_detection.x)
        yA = max(self.y, other_detection.y)
        xB = min(self.x + self.width, other_detection.x + other_detection.width)
        yB = min(self.y + self.height, other_detection.y + other_detection.height)
        inter_area = max(xB - xA, 0) * max(yB - yA, 0)
        union = self.area + other_detection.area - inter_area
        if union == 0:
            print('WARNING: Something went wrong while calculating the intersection over union. Returning 0.')
            return 0
        return inter_area / union

    @property
    def area(self) -> float:
        return self.width * self.height


@dataclass(slots=True, kw_only=True)
class Shape:
    points: list[Point]

    @staticmethod
    def from_str(points: str) -> Shape:
        p = points.split(',')
        return Shape(points=[Point(x=float(p[i]), y=float(p[i+1])) for i in range(0, len(p), 2)])

    def __str__(self) -> str:
        return ', '. join([f'x:{p.x:.0f} y:{p.y:.0f}' for p in self.points])


@dataclass(slots=True, kw_only=True)
class SegmentationDetection:
    category_name: str
    model_name: str
    confidence: float
    shape: Shape

    def __str__(self) -> str:
        return f'shape:{self.shape.__str__} cat:{self.category_name} conf:{self.confidence:.2f}'

    def to_svg(self, shrink: int = 1) -> str:
        color = 'green'
        d = ' '.join([f"{'L' if i > 0 else 'M'} {p.x/shrink} {p.y/shrink}" for i, p in enumerate(self.shape.points)])
        d += ' Z'
        p0 = self.shape.points[0]
        return (
            f'<path d="{d}" stroke-width="2" stroke="{color}" fill="none" />'
            f'<text x="{p0.x / shrink + 10}" y="{p0.y / shrink + 4}" text-anchor="start" stroke="{color}" '
            f'fill="{color}" font-size="12" font-weight="light">{self.category_name} ({self.confidence:.0%})</text>'
        )


@dataclass(slots=True, kw_only=True)
class PointDetection(Detection):

    @property
    def cx(self) -> float:
        return self.x

    @property
    def cy(self) -> float:
        return self.y

    def distance(self, other: PointDetection) -> float:
        return np.sqrt((other.x - self.x)**2 + (other.y - self.y)**2)

    def __str__(self) -> str:
        return f'x:{self.x:.0f} y:{self.y:.0f} cat:{self.category_name} conf:{self.confidence:.2f}'

    def to_svg(self, shrink: int = 1) -> str:
        color = 'red'
        x = self.x / shrink
        y = self.y / shrink
        return (
            f'<circle cx="{x}" cy="{y}" r="4" stroke-width="2" stroke="{color}" fill="none" />'
            f'<text x="{x + 10}" y="{y + 4}" text-anchor="start" stroke="{color}" fill="{color}" font-size="12" '
            f'font-weight="light">{self.category_name} ({self.confidence:.0%})</text>'
        )


@dataclass(slots=True, kw_only=True)
class Detections:
    boxes: list[BoxDetection] = field(default_factory=list)
    points: list[PointDetection] = field(default_factory=list)
    segmentations: list[SegmentationDetection] = field(default_factory=list)

    def to_svg(self, shrink: int = 1) -> str:
        return '\n'.join(
            b.to_svg(shrink) for b in self.boxes) + '\n' + '\n'.join(
            p.to_svg(shrink) for p in self.points) + '\n' + '\n'.join(
            s.to_svg(shrink) for s in self.segmentations)

    def to_dict(self) -> dict:
        return asdict(self)



import logging
from typing import Any

from typing_extensions import Self

from ... import rosys
from ..camera import ConfigurableCamera, TransformableCamera
from ..image import Image
from ..image_processing import get_image_size_from_bytes, process_jpeg_image
from ..image_rotation import ImageRotation
from .mjpeg_device import MjpegDevice
from .mjpeg_device_factory import MjpegDeviceFactory


class MjpegCamera(TransformableCamera, ConfigurableCamera):

    def __init__(self,
                 *,
                 id: str,  # pylint: disable=redefined-builtin
                 name: str | None = None,
                 connect_after_init: bool = True,
                 base_path_overwrite: str | None = None,
                 username: str | None = None,
                 password: str | None = None,
                 ip: str | None = None,
                 fps: int = 10,
                 resolution: tuple[int, int] = (640, 480),
                 mirrored: bool = False,
                 **kwargs: Any,
                 ) -> None:
        super().__init__(id=id, name=name, connect_after_init=connect_after_init,
                         base_path_overwrite=base_path_overwrite, **kwargs)
        self.log = logging.getLogger(f'rosys.vision.mjpeg_camera.{self.id}')
        self.username = username
        self.password = password

        self.ip = ip

        self.index: int | None = None
        parts = self.id.split('-')
        if len(parts) == 2 and parts[1].isdigit():
            self.index = int(parts[1])

        self.mac = parts[0]
        self.device: MjpegDevice | None = None

        self._register_parameter('fps', self._get_fps, self._set_fps, default_value=fps)
        self._register_parameter('resolution', self._get_resolution, self._set_resolution, default_value=resolution)
        self._register_parameter('mirrored', self._get_mirrored, self._set_mirrored, default_value=mirrored)

    def to_dict(self) -> dict:
        return super().to_dict() | {
            name: param.value for name, param in self._parameters.items()
        } | {
            'username': self.username,
            'password': self.password,
            'ip': self.ip,
        }

    @classmethod
    def from_dict(cls, data: dict) -> Self:
        return cls(**data)

    @property
    def is_connected(self) -> bool:
        return (self.device is not None) and self.device.is_connected

    async def connect(self) -> None:
        if self.is_connected:
            return

        if not self.ip:
            self.log.error('No IP address provided')
            return

        try:
            self.device = MjpegDeviceFactory.create(self.mac, self.ip, index=self.index, username=self.username,
                                                    password=self.password, on_new_image_data=self._handle_new_image_data)
        except ValueError as error:
            self.log.error('Could not connect to device: %s', error)
            return

        await self._apply_all_parameters()

    async def disconnect(self) -> None:
        if self.device is None:
            return

        self.device.shutdown()
        self.device = None

    async def _handle_new_image_data(self, image: bytes, timestamp: float) -> None:
        if self.crop or self.rotation != ImageRotation.NONE:
            image_ = await rosys.run.cpu_bound(process_jpeg_image, image, self.rotation, self.crop)
            if image_ is None:
                return
            image = image_
        try:
            final_image_resolution = get_image_size_from_bytes(image)
        except ValueError:
            return

        self._add_image(Image(camera_id=self.id, data=image, time=timestamp, size=final_image_resolution))

    async def _set_fps(self, fps: int) -> None:
        assert self.device is not None
        await self.device.set_fps(fps)

    async def _get_fps(self) -> int:
        assert self.device is not None

        return await self.device.get_fps()

    async def _set_resolution(self, resolution: tuple[int, int]) -> None:
        assert self.device is not None

        await self.device.set_resolution(*resolution)

    async def _get_resolution(self) -> tuple[int, int]:
        assert self.device is not None

        return await self.device.get_resolution()

    async def _set_mirrored(self, mirrored: bool) -> None:
        assert self.device is not None

        await self.device.set_mirrored(mirrored)

    async def _get_mirrored(self) -> bool:
        assert self.device is not None

        return await self.device.get_mirrored()



import asyncio
from collections.abc import Coroutine
from typing import Any, TypeVar

_T = TypeVar('_T')


class LazyWorker:

    def __init__(self) -> None:
        self._condition = asyncio.Event()
        self._is_free = True

    async def run(self, coro: Coroutine[Any, None, _T]) -> _T | None:
        """Run the coroutine and return the result.

        If the worker is busy, the coroutine is put into a queue of size 1.
        All other coroutines are discarded.

        :return: result of the coroutine or ``None`` if the coroutine was discarded
        """
        self._notify()

        if not self._is_free:
            await self._condition.wait()

        if not self._is_free:
            coro.close()
            return None

        self._is_free = False
        try:
            return await coro
        finally:
            self._is_free = True
            self._notify()

    def _notify(self) -> None:
        self._condition.set()
        self._condition.clear()



from enum import Enum


class VendorType(Enum):
    AXIS = 1
    MOTEC = 2
    OTHER = -1


def mac_to_vendor(mac: str) -> VendorType:
    if mac.startswith(('00:40:8c', 'ac:cc:8e', 'b8:a4:4f', 'e8:27:25')):
        return VendorType.AXIS
    if mac.startswith('2c:26:5f'):
        return VendorType.MOTEC
    return VendorType.OTHER


def mac_to_url(mac: str, ip: str, *, index: int | None = None) -> str | None:
    vendor = mac_to_vendor(mac)
    if vendor == VendorType.AXIS:
        return f'http://{ip}/axis-cgi/mjpg/video.cgi?' + (f'camera={index}' if index is not None else '')
    if vendor == VendorType.MOTEC:
        return f'http://{ip}:1001/stream.mjpg'
    return None



import asyncio
import uuid
from collections import defaultdict
from collections.abc import Callable, Coroutine
from functools import wraps
from typing import ParamSpec, TypeVar

from nicegui import ui

_T = TypeVar('_T')
_P = ParamSpec('_P')


class Track:

    def __init__(self) -> None:
        self.stacks: defaultdict[int, dict[int, str]] = defaultdict(dict)

    def __call__(self, f: Callable[_P, Coroutine[None, None, _T]]) -> Callable[_P, Coroutine[None, None, _T]]:
        @wraps(f)
        async def wrap(*args, **kwargs):
            task_id = id(asyncio.current_task())
            stack = self.stacks[task_id]
            uid = uuid.uuid4().int
            stack[uid] = f.__name__
            try:
                return await f(*args, **kwargs)
            finally:
                del stack[uid]
                if not stack:
                    del self.stacks[task_id]
        return wrap

    def ui(self) -> ui.markdown:
        markdown = ui.markdown()
        ui.timer(0.1, lambda: markdown.set_content('\n\n'.join(' → '.join(stack.values()).replace('_', r'\_')
                                                               for stack in self.stacks.values())))
        return markdown


track = Track()



import logging

from .camera import Camera
from .camera_provider import CameraProvider


class MultiCameraProvider(CameraProvider):
    """A multi-camera provider combines multiple camera providers into one.

    This is useful if another module requires a single camera provider but the robot has multiple camera sources like USB and WiFi cameras.
    """

    def __init__(self, *camera_providers: CameraProvider) -> None:
        super().__init__()
        self.providers = camera_providers
        self.log = logging.getLogger('rosys.multi_camera_provider')
        for camera_provider in self.providers:
            camera_provider.NEW_IMAGE.register(self.NEW_IMAGE.emit)
            camera_provider.CAMERA_ADDED.register(self.CAMERA_ADDED.emit)
            camera_provider.CAMERA_REMOVED.register(self.CAMERA_REMOVED.emit)

    async def update_device_list(self) -> None:
        for provider in self.providers:
            try:
                await provider.update_device_list()
            except Exception as e:
                self.log.error('Error while scanning for cameras in "%s": %s', provider.__class__.__name__, e)

    def request_backup(self) -> None:
        for provider in self.providers:
            provider.request_backup()

    @property
    def cameras(self) -> dict[str, Camera]:
        return {id: camera for provider in self.providers for id, camera in provider.cameras.items()}

    def remove_camera(self, camera_id: str) -> None:
        for provider in self.providers:
            if camera_id in provider.cameras:
                provider.remove_camera(camera_id)

    def add_camera(self, camera) -> None:
        raise NotImplementedError('Adding cameras to a MultiCameraProvider is not supported')



import warnings
from collections.abc import Callable
from functools import wraps


def deprecated_param(param_name: str, *, remove_in_version: str | None = None, stacklevel: int = 2) -> Callable:
    """Mark a function parameter as deprecated."""
    def decorator(func: Callable) -> Callable:
        @wraps(func)
        def wrapper(*args, **kwargs):
            if param_name in kwargs:
                warnings.warn(
                    f'The parameter "{param_name}" is deprecated and will be removed in '
                    f'{("RoSys " + remove_in_version) if remove_in_version else "a future version"}.',
                    category=DeprecationWarning,
                    stacklevel=stacklevel,
                )
            return func(*args, **kwargs)
        return wrapper
    return decorator


def deprecated_function(*, remove_in_version: str | None = None, stacklevel: int = 2) -> Callable:
    """Mark a function as deprecated."""
    def decorator(func: Callable) -> Callable:
        @wraps(func)
        def wrapper(*args, **kwargs):
            warnings.warn(
                f'The function "{func.__name__}" is deprecated and will be removed in '
                f'{("RoSys " + remove_in_version) if remove_in_version else "a future version"}.',
                category=DeprecationWarning,
                stacklevel=stacklevel,
            )
            return func(*args, **kwargs)
        return wrapper
    return decorator



from __future__ import annotations

import logging
import math
from abc import ABC
from dataclasses import dataclass
from enum import IntEnum

import numpy as np
import serial
from nicegui import ui
from serial.tools import list_ports

from .. import rosys
from ..driving.driver import PoseProvider
from ..event import Event
from ..geometry import GeoPoint, GeoPose, Pose
from ..run import io_bound


class GpsQuality(IntEnum):
    INVALID = 0
    GPS = 1
    DGPS = 2
    PPS = 3
    RTK_FIXED = 4
    RTK_FLOAT = 5
    DEAD_RECKONING = 6
    MANUAL = 7
    SIMULATION = 8


@dataclass(slots=True, kw_only=True)
class GnssMeasurement:
    time: float
    pose: GeoPose
    latitude_std_dev: float = 0.0
    longitude_std_dev: float = 0.0
    heading_std_dev: float = 0.0
    gps_quality: GpsQuality = GpsQuality.INVALID
    num_satellites: int = 0
    hdop: float = 0.0
    altitude: float = 0.0

    @property
    def point(self) -> GeoPoint:
        return self.pose.point

    @property
    def heading(self) -> float:
        return self.pose.heading


class Gnss(ABC):
    """A GNSS module that provides measurements from a GNSS receiver."""

    def __init__(self) -> None:
        self.log = logging.getLogger('rosys.gnss')
        self.last_measurement: GnssMeasurement | None = None

        self.NEW_MEASUREMENT = Event()
        """a new measurement has been received (argument: ``GnssMeasurement``)"""

    @property
    def is_connected(self) -> bool:
        return False

    def developer_ui(self) -> None:
        ui.label('GNSS').classes('text-center text-bold')
        with ui.grid(columns='auto auto').classes('gap-y-1 w-2/5'):
            ui.label('Connected:')
            ui.icon('close').bind_name_from(self, 'is_connected',
                                            lambda x: 'check' if x else 'close')
            ui.label('Position:')
            with ui.column().classes('gap-y-0'):
                ui.label().bind_text_from(self, 'last_measurement',
                                          lambda x: f'{math.degrees(x.pose.lat):.8f}˚' if x else '-')
                ui.label().bind_text_from(self, 'last_measurement',
                                          lambda x: f'± {x.latitude_std_dev:.3f}m' if x else '-')
                ui.label().bind_text_from(self, 'last_measurement',
                                          lambda x: f'{math.degrees(x.pose.lon):.8f}˚' if x else '-')
                ui.label().bind_text_from(self, 'last_measurement',
                                          lambda x: f'± {x.longitude_std_dev:.3f}m' if x else '-')
            ui.label('Heading:')
            with ui.column().classes('gap-y-0'):
                ui.label().bind_text_from(self, 'last_measurement',
                                          lambda x: f'{math.degrees(x.pose.heading):.2f}˚' if x else '-')
                ui.label().bind_text_from(self, 'last_measurement',
                                          lambda x: f'± {x.heading_std_dev:.2f}˚' if x else '-')
            ui.label('Quality:')
            ui.label().bind_text_from(self, 'last_measurement',
                                      lambda x: x.gps_quality.name if x else '')
            ui.label('Satellites:')
            ui.label().bind_text_from(self, 'last_measurement',
                                      lambda x: str(x.num_satellites) if x else '-')
            ui.label('HDOP:')
            ui.label().bind_text_from(self, 'last_measurement',
                                      lambda x: f'{x.hdop:.2f}' if x else '-')
            ui.label('Altitude:')
            ui.label().bind_text_from(self, 'last_measurement',
                                      lambda x: f'{x.altitude:.3f}m' if x else '-')
            ui.label('Last update:')
            ui.label().bind_text_from(self, 'last_measurement',
                                      lambda x: f'{rosys.time() - x.time:.2f}s' if x else '')


class GnssHardware(Gnss):
    """This hardware module connects to a Septentrio SimpleRTK3b (Mosaic-H) GNSS receiver."""

    def __init__(self, *, antenna_pose: Pose | None) -> None:
        """
        :param antenna_pose: the pose of the main antenna in the robot's coordinate frame (yaw: direction to the auxiliary antenna)
        """
        super().__init__()
        self.antenna_pose = antenna_pose or Pose(x=0.0, y=0.0, yaw=0.0)
        serial_device_path = self._find_device()
        self.serial_connection = self._connect_to_device(serial_device_path)
        rosys.on_startup(self._run)

    @property
    def is_connected(self) -> bool:
        if self.serial_connection is None:
            self.log.debug('Device not connected')
            return False
        if not self.serial_connection.isOpen():
            self.log.debug('Device not open')
            return False
        return True

    async def _run(self) -> None:
        buffer = ''
        last_raw_latitude = 0.0
        last_raw_longitude = 0.0
        last_raw_heading = 0.0
        last_latitude_accuracy = 0.0
        last_longitude_accuracy = 0.0
        last_heading_accuracy = 0.0
        last_gps_quality = 0
        last_num_satellites = 0
        last_hdop = 0.0
        last_altitude = 0.0
        last_gga_timestamp = ''
        last_gst_timestamp = ''
        last_pssn_timestamp = ''
        while True:
            if not self.is_connected:
                return None
            result = await io_bound(self.serial_connection.read_until, b'\r\n')
            if not result:
                self.log.debug('No data')
                continue
            line = result.decode('utf-8')
            if line.endswith('\r\n'):
                sentence = buffer + line.split('*')[0]
                buffer = ''
            else:
                buffer += line
                continue

            self.log.debug(sentence)
            parts = sentence.split(',')
            try:
                timestamp = parts[{'$GPGGA': 1, '$GPGST': 1, '$PSSN': 2}[parts[0]]]
                if parts[0] == '$GPGGA':
                    if parts[2] == '' or parts[4] == '':
                        continue
                    last_gga_timestamp = timestamp
                    last_raw_latitude = self._convert_to_decimal(parts[2], parts[3])
                    last_raw_longitude = self._convert_to_decimal(parts[4], parts[5])
                    last_gps_quality = int(parts[6]) if parts[6] else 0
                    last_num_satellites = int(parts[7]) if parts[7] else 0
                    last_hdop = float(parts[8]) if parts[8] else 0.0
                    last_altitude = float(parts[9]) if parts[9] else 0.0
                if parts[0] == '$GPGST' and parts[6] and parts[7]:
                    last_gst_timestamp = timestamp
                    last_latitude_accuracy = float(parts[6])
                    last_longitude_accuracy = float(parts[7])
                if parts[0] == '$PSSN' and parts[1] == 'HRP':
                    last_pssn_timestamp = timestamp
                    last_raw_heading = float(parts[4] or 0.0)
                    last_heading_accuracy = float(parts[7] or 'inf')
                if last_gga_timestamp == last_gst_timestamp == last_pssn_timestamp != '':
                    last_heading = last_raw_heading + self.antenna_pose.yaw_deg
                    antenna_pose = GeoPose.from_degrees(last_raw_latitude, last_raw_longitude, last_heading)
                    robot_pose = antenna_pose.relative_shift_by(x=-self.antenna_pose.x, y=-self.antenna_pose.y)
                    self.last_measurement = GnssMeasurement(
                        time=rosys.time(),
                        pose=robot_pose,
                        latitude_std_dev=last_latitude_accuracy,
                        longitude_std_dev=last_longitude_accuracy,
                        heading_std_dev=last_heading_accuracy,
                        gps_quality=GpsQuality(last_gps_quality),
                        num_satellites=last_num_satellites,
                        hdop=last_hdop,
                        altitude=last_altitude,
                    )
                    self.NEW_MEASUREMENT.emit(self.last_measurement)
            except Exception as e:
                self.log.exception(e)

    # TODO: move to helper and add search argument; for other serial devices
    def _find_device(self) -> str:
        for port in list_ports.comports():
            self.log.debug('Found port: %s - %s', port.device, port.description)
            if 'Septentrio' in port.description:
                self.log.info('Found GNSS device: %s', port.device)
                return port.device
        raise RuntimeError('No GNSS device found')

    def _connect_to_device(self, port: str, *, baudrate: int = 115200, timeout: float = 0.2) -> serial.Serial:
        self.log.info('Connecting to GNSS device "%s"...', port)
        try:
            return serial.Serial(port=port, baudrate=baudrate, timeout=timeout)
        except serial.SerialException as e:
            raise RuntimeError(f'Could not connect to GNSS device: {port}') from e

    @staticmethod
    def _convert_to_decimal(coord: str, direction: str) -> float:
        """Convert a coordinate in the format DDMM.mmmm to decimal degrees.

        :param coord: the coordinate to convert
        :param direction: the direction (N/S/E/W)
        :return: the coordinate in decimal degrees
        """
        split_index = coord.find('.') - 2
        degrees = float(coord[:split_index])
        minutes = float(coord[split_index:])
        decimal = degrees + minutes / 60
        if direction in ['S', 'W']:
            decimal = -decimal
        return decimal


class GnssSimulation(Gnss):
    """Simulation of a GNSS receiver."""

    def __init__(self, *,
                 wheels: PoseProvider,
                 lat_std_dev: float = 0.01,
                 lon_std_dev: float = 0.01,
                 heading_std_dev: float = 0.01,
                 gps_quality: GpsQuality = GpsQuality.RTK_FIXED) -> None:
        """
        :param wheels: the wheels to use for the simulation
        :param lat_std_dev: the standard deviation of the latitude in meters
        :param lon_std_dev: the standard deviation of the longitude in meters
        :param heading_std_dev: the standard deviation of the heading in degrees
        :param gps_quality: the quality of the GPS signal
        """
        super().__init__()
        self.wheels = wheels
        self._is_connected = True
        self._lat_std_dev = lat_std_dev
        self._lon_std_dev = lon_std_dev
        self._heading_std_dev = heading_std_dev
        self._gps_quality = gps_quality
        self.last_measurement: GnssMeasurement | None = None
        rosys.on_repeat(self.simulate, 1.0)

    @property
    def is_connected(self) -> bool:
        return self._is_connected

    @is_connected.setter
    def is_connected(self, value: bool) -> None:
        self._is_connected = value

    def simulate(self) -> None:
        if not self.is_connected:
            return
        geo_pose = GeoPose.from_pose(self.wheels.pose)

        noise_lat = np.random.normal(0, self._lat_std_dev)
        noise_lon = np.random.normal(0, self._lon_std_dev)
        noise_heading = np.random.normal(0, math.radians(self._heading_std_dev))

        direction = math.atan2(noise_lon, noise_lat)
        noise_point = geo_pose.point.polar(noise_lat, direction)
        noise_pose = GeoPose(lat=noise_point.lat, lon=noise_point.lon, heading=geo_pose.heading + noise_heading)

        self.last_measurement = GnssMeasurement(
            time=rosys.time(),
            pose=noise_pose,
            latitude_std_dev=self._lat_std_dev,
            longitude_std_dev=self._lon_std_dev,
            heading_std_dev=self._heading_std_dev,
            gps_quality=self._gps_quality,
        )
        self.NEW_MEASUREMENT.emit(self.last_measurement)

    def developer_ui(self) -> None:
        super().developer_ui()
        ui.label('Simulation').classes('text-center text-bold')
        with ui.column().classes('gap-y-1'):
            ui.checkbox('Connected').bind_value(self, '_is_connected')
            ui.number(label='Latitude Std Dev', format='%.3f', prefix='± ', suffix='m') \
                .bind_value(self, '_lat_std_dev').classes('w-4/5')
            ui.number(label='Longitude Std Dev', format='%.3f', prefix='± ', suffix='m') \
                .bind_value(self, '_lon_std_dev').classes('w-4/5')
            ui.number(label='Heading Std Dev', format='%.2f', prefix='± ', suffix='°') \
                .bind_value(self, '_heading_std_dev').classes('w-4/5')
            ui.select({quality: quality.name for quality in GpsQuality}, label='Quality') \
                .bind_value(self, '_gps_quality').classes('w-4/5')



import asyncio
from collections.abc import Coroutine


class parallelize:
    """Parallelize multiple coroutines.

    This class allows to combine multiple coroutines into one that can be passed to the
    `automator <https://rosys.io/reference/rosys/automation/#rosys.automation.Automator>`__
    to run them in parallel.
    """

    def __init__(self, *coros: Coroutine, return_when_first_completed: bool = False) -> None:
        self.coros = coros
        self.return_when_first_completed = return_when_first_completed

    def __await__(self):
        coro_iters = [coro.__await__() for coro in self.coros]
        iter_sends = [coro_iter.send for coro_iter in coro_iters]
        iter_throws = [coro_iter.throw for coro_iter in coro_iters]
        sends = iter_sends[:]
        messages = [None for _ in sends]
        completed = [False for _ in sends]
        waiting: list[asyncio.Future | asyncio.Task | None] = [None for _ in sends]

        try:
            while not all(completed):
                active_coros = [i for i, is_completed in enumerate(completed) if not is_completed]

                for i in active_coros:
                    task = waiting[i]
                    if task is not None:
                        if task.done():
                            waiting[i] = None
                        else:
                            continue

                    try:
                        signal = sends[i](messages[i])
                    except StopIteration:
                        completed[i] = True
                        if self.return_when_first_completed:
                            return
                        continue
                    else:
                        sends[i] = iter_sends[i]

                    if isinstance(signal, asyncio.Future | asyncio.Task):
                        waiting[i] = signal
                    else:
                        try:
                            messages[i] = yield signal
                        except GeneratorExit:
                            raise
                        except BaseException as e:
                            try:
                                sends[i] = iter_throws[i](type(e), e, e.__traceback__)
                            except StopIteration:
                                completed[i] = True
                            else:
                                sends[i] = iter_sends[i]  # Reset to send method if throw doesn't raise StopIteration

                if all(waiting[i] is not None or completed[i] for i in active_coros):
                    yield  # Allow event loop to process other tasks
        finally:
            # Ensure all coroutines are properly closed
            close_exceptions = []
            for i, is_completed in enumerate(completed):
                if not is_completed:
                    try:
                        coro_iters[i].close()
                    except Exception as e:
                        close_exceptions.append((i, e))

            # If any exceptions occurred during closing, raise a RuntimeError with details
            if close_exceptions:
                error_msg = 'Exceptions occurred while closing coroutines:\n'
                error_msg += '\n'.join(f'Coroutine {i}: {e}' for i, e in close_exceptions)
                raise RuntimeError(error_msg)



import numpy as np
from typing_extensions import Self

from ...geometry import Frame3d, Pose3d, Rotation
from ...persistence.converters import from_dict, to_dict
from ..calibration import Calibration, Intrinsics
from .camera import Camera


class CalibratableCamera(Camera):

    def __init__(self, calibration: Calibration | None = None, **kwargs) -> None:
        super().__init__(**kwargs)
        self.calibration = calibration

    def to_dict(self) -> dict:
        return super().to_dict() | {'calibration': to_dict(self.calibration)}

    @classmethod
    def args_from_dict(cls, data: dict) -> dict:
        calibration_data = data.pop('calibration', None)
        calibration = from_dict(Calibration, calibration_data) if calibration_data else None
        return {'calibration': calibration, **super().args_from_dict(data)}

    @property
    def is_calibrated(self) -> bool:
        return self.calibration is not None

    @classmethod
    def create_calibrated(cls, *,
                          width: int = 800, height: int = 600, focal_length: float = 570,
                          x: float = 0.0, y: float = 0.0, z: float = 1.0,
                          roll: float = np.pi, pitch: float = 0.0, yaw: float = 0.0,
                          frame: Frame3d | None = None,
                          **kwargs) -> Self:
        camera = cls(**kwargs)
        camera.set_perfect_calibration(width=width, height=height, focal_length=focal_length,
                                       x=x, y=y, z=z,
                                       roll=roll, pitch=pitch, yaw=yaw,
                                       frame=frame)
        return camera

    def set_perfect_calibration(self, *,
                                width: int = 800, height: int = 600, focal_length: float = 570,
                                x: float = 0.0, y: float = 0.0, z: float = 1.0,
                                roll: float = np.pi, pitch: float = 0.0, yaw: float = 0.0,
                                frame: Frame3d | None = None,
                                ) -> None:
        self.calibration = Calibration(
            intrinsics=Intrinsics.create_default(width, height, focal_length=focal_length),
            extrinsics=Pose3d(x=x, y=y, z=z, rotation=Rotation.from_euler(roll, pitch, yaw)).in_frame(frame),
        )



from __future__ import annotations

from dataclasses import dataclass
from typing import overload

import numpy as np

from .. import helpers
from .line import Line
from .point import Point
from .point3d import Point3d


@dataclass(slots=True, kw_only=True)
class PoseStep:
    linear: float
    angular: float
    time: float


@dataclass(slots=True, kw_only=True)
class Pose:
    x: float = 0
    y: float = 0
    yaw: float = 0
    time: float = 0

    @property
    def point(self) -> Point:
        return Point(x=self.x, y=self.y)

    def point_3d(self, z: float = 0) -> Point3d:
        return Point3d(x=self.x, y=self.y, z=z)

    @property
    def matrix(self) -> np.ndarray:
        return np.array([
            [np.cos(self.yaw), -np.sin(self.yaw), self.x],
            [np.sin(self.yaw), np.cos(self.yaw), self.y],
            [0, 0, 1],
        ])

    @property
    def inv_matrix(self) -> np.ndarray:
        return np.array([
            [np.cos(self.yaw), np.sin(self.yaw), -np.cos(self.yaw) * self.x - np.sin(self.yaw) * self.y],
            [-np.sin(self.yaw), np.cos(self.yaw), np.sin(self.yaw) * self.x - np.cos(self.yaw) * self.y],
            [0, 0, 1],
        ])

    @property
    def yaw_deg(self) -> float:
        return np.rad2deg(self.yaw)

    @staticmethod
    def from_matrix(M, time: float = 0) -> Pose:
        return Pose(x=M[0, -1], y=M[1, -1], yaw=np.arctan2(M[1, 0], M[0, 0]), time=time)

    def __str__(self) -> str:
        return f'{self.x:.3f}, {self.y:.3f}, {self.yaw_deg:.1f} deg'

    @overload
    def distance(self, other: Point) -> float: ...

    @overload
    def distance(self, other: Pose) -> float: ...

    def distance(self, other: Point | Pose) -> float:
        return float(np.sqrt((other.x - self.x)**2 + (other.y - self.y)**2))

    def projected_distance(self, other: Pose) -> float:
        return self.point.projected_distance(other.point, other.yaw)

    @overload
    def direction(self, other: Point) -> float: ...

    @overload
    def direction(self, other: Pose) -> float: ...

    def direction(self, other: Point | Pose) -> float:
        return float(np.arctan2(other.y - self.y, other.x - self.x))

    @overload
    def relative_direction(self, other: Point) -> float: ...

    @overload
    def relative_direction(self, other: Pose) -> float: ...

    def relative_direction(self, other: Point | Pose) -> float:
        return helpers.angle(self.yaw, self.direction(other))

    def __iadd__(self, step: PoseStep) -> Pose:
        self.x += step.linear * np.cos(self.yaw)
        self.y += step.linear * np.sin(self.yaw)
        self.yaw += step.angular
        self.time = step.time
        return self

    def __add__(self, step: PoseStep) -> Pose:
        return Pose(
            x=self.x + step.linear * np.cos(self.yaw),
            y=self.y + step.linear * np.sin(self.yaw),
            yaw=self.yaw + step.angular,
            time=step.time,
        )

    def transform(self, point: Point) -> Point:
        return Point(
            x=self.x + point.x * np.cos(self.yaw) - point.y * np.sin(self.yaw),
            y=self.y + point.x * np.sin(self.yaw) + point.y * np.cos(self.yaw),
        )

    def transform3d(self, point: Point3d) -> Point3d:
        return Point3d(
            x=self.x + point.x * np.cos(self.yaw) - point.y * np.sin(self.yaw),
            y=self.y + point.x * np.sin(self.yaw) + point.y * np.cos(self.yaw),
            z=point.z,
        )

    def transform_array(self, points: np.ndarray) -> np.ndarray:
        return np.stack((
            self.x + points[:, 0] * np.cos(self.yaw) - points[:, 1] * np.sin(self.yaw),
            self.y + points[:, 0] * np.sin(self.yaw) + points[:, 1] * np.cos(self.yaw),
        ), axis=1)

    def transform_pose(self, pose: Pose) -> Pose:
        return Pose(
            x=self.x + pose.x * np.cos(self.yaw) - pose.y * np.sin(self.yaw),
            y=self.y + pose.x * np.sin(self.yaw) + pose.y * np.cos(self.yaw),
            yaw=self.yaw + pose.yaw,
            time=pose.time,
        )

    def transform_line(self, line: Line) -> Line:
        a, b, c = self.inv_matrix.T @ line.tuple
        return Line(a=a, b=b, c=c)

    def relative_pose(self, other: Pose) -> Pose:
        return Pose.from_matrix(self.inv_matrix @ other.matrix)

    def relative_point(self, other: Point) -> Point:
        dx = other.x - self.x
        dy = other.y - self.y
        return Point(x=dx * np.cos(-self.yaw) - dy * np.sin(-self.yaw),
                     y=dx * np.sin(-self.yaw) + dy * np.cos(-self.yaw))

    def rotate(self, angle: float) -> Pose:
        return Pose(x=self.x, y=self.y, yaw=self.yaw+angle, time=self.time)

    def interpolate(self, other: Pose, t: float) -> Pose:
        return Pose(
            x=(1 - t) * self.x + t * other.x,
            y=(1 - t) * self.y + t * other.y,
            yaw=(1 - t) * self.yaw + t * other.yaw,
            time=(1 - t) * self.time + t * other.time,
        )



import json
import time
from dataclasses import dataclass
from typing import Any

import httpx
from httpx import QueryParams


@dataclass
class JovisionCameraSettings:
    """Dataclass that holds the settings of an RTSP camera"""
    codec: str
    fps: int
    quality: str
    resolution: tuple[int, int]
    bitrate_control: str
    bitrate: int
    stream_id: int
    channel_id: int = 0

    def to_dict(self) -> dict[str, Any]:
        return {
            'channelid': self.channel_id,
            'streamid': self.stream_id,
            'venctype': self.codec,
            'width': self.resolution[0],
            'height': self.resolution[1],
            'frameRate': self.fps,
            'bitRate': self.bitrate,
            'ngop_s': 100,
            'quality': self.quality,
            'rcMode': self.bitrate_control,
        }


class JovisionInterface:
    """Interface that communicates with the RTSP camera settings mimicking the webpage requests"""

    def __init__(self, ip: str) -> None:
        self.ip = ip

    @property
    def settings_url(self) -> str:
        return f'http://{self.ip}/cgi-bin/jvsweb.cgi'

    async def _send_settings(self, settings: list[JovisionCameraSettings]) -> None:
        """
        Set the settings for all streams
        """
        streams: list[dict] = []
        for i, setting in enumerate(settings):
            parameter_dict = setting.to_dict()
            parameter_dict['smartencode'] = 'open' if i < len(settings) - 1 else 'close'
            streams.append(parameter_dict)

        cmd = {
            'method': 'stream_set_params',
            'user': {
                'name': 'admin',
                'digest': '479ead4555b49227dc8812d06970c652',
            },
            'param': {
                'streams': streams,
            },
        }
        params = QueryParams(
            cmd=json.dumps(cmd),
            _=int(time.time() * 1000),  # current time as a timestamp
        )
        async with httpx.AsyncClient() as client:
            await client.get(self.settings_url, params=params)

    async def set_fps(self, stream_id: int, fps: int) -> None:
        current_settings = await self.get_current_settings()
        for settings in current_settings:
            if settings.stream_id == stream_id:
                settings.fps = int(fps)
                break
        await self._send_settings(current_settings)

    async def get_fps(self, stream_id: int) -> int | None:
        current_settings = await self.get_current_settings()
        for settings in current_settings:
            if settings.stream_id == stream_id:
                return settings.fps

        return None

    async def get_bitrate(self, stream_id: int) -> int | None:
        current_settings = await self.get_current_settings()
        for settings in current_settings:
            if settings.stream_id == stream_id:
                return settings.bitrate

        return None

    async def set_bitrate(self, stream_id: int, bitrate: int) -> None:
        current_settings = await self.get_current_settings()
        for settings in current_settings:
            if settings.stream_id == stream_id:
                settings.bitrate = int(bitrate)
                break
        await self._send_settings(current_settings)

    async def get_current_settings(self) -> list[JovisionCameraSettings]:
        cmd = {
            'method': 'stream_get_params',
            'user': {
                'name': 'admin',
                'digest': 'd065077d3e61cb56d45bd2dd28b83842',
            },
            'param': {
                'channelid': 0,
            },
        }
        params = QueryParams(
            cmd=json.dumps(cmd),
            _=int(time.time() * 1000),
        )
        async with httpx.AsyncClient() as client:
            response = await client.get(self.settings_url, params=params)

        return [
            JovisionCameraSettings(
                codec=stream['venctype'],
                fps=stream['frameRate'],
                quality=stream['quality'],
                resolution=(stream['width'], stream['height']),
                bitrate_control=stream['rcMode'],
                bitrate=stream['bitRate'],
                stream_id=stream['streamid'],
                channel_id=stream['channelid'],
            )
            for stream in response.json()['result']['streams']
        ]

    async def get_parameter_ranges(self):
        raise NotImplementedError

        # cmd = {
        #     'method': 'stream_get_all_ability',
        #     'user': {
        #         'name': 'admin',
        #         'digest': '8d5985ea2b9994aacb6cb3e3f826aae5'
        #     },
        #     'param': {
        #         'channelid': 0,
        #     }
        # }
        # params = {
        #     'cmd': json.dumps(cmd),
        #     '_': time.time() * 1000,
        # }
        # async with httpx.AsyncClient() as client:
        #     response = await client.get(self.settings_url, params=params)

        # for stream_id, stream in enumerate(response.json()['result']['all']):
        #     print(f'stream {stream_id}')
        #     resolutions = stream['resolutions']
        #     for resolution in resolutions:
        #         width = resolution['width']
        #         height = resolution['height']
        #         b_default = resolution['bDefault']
        #         max_kbps = resolution['maxKbps']
        #         min_kbps = resolution['minKbps']
        #         def_kbps = resolution['defKbps']
        #         max_fr = resolution['maxFr']
        #         min_fr = resolution['minFr']
        #         def_fr = resolution['defFr']
        #     max_quality = stream['maxQuality']
        #     min_quality = stream['minQuality']
        #     max_ngop = stream['maxNGOP']
        #     min_ngop = stream['minNGOP']
        #     b_support_h265 = stream['bSupportH265']
        #     b_support_smart_enc = stream['bSupportSmartEnc']



from dataclasses import dataclass, field
from datetime import datetime
from typing import Literal
from uuid import uuid4

import numpy as np

from .. import rosys
from ..geometry import Point3d
from .calibratable_camera_provider import CalibratableCameraProvider
from .camera import CalibratableCamera
from .detections import BoxDetection, Detections, PointDetection
from .detector import Autoupload, Detector, DetectorInfo, ModelVersioningInfo
from .image import Image


@dataclass(slots=True, kw_only=True)
class SimulatedObject:
    category_name: str
    position: Point3d
    size: tuple[float, float, float] | None = None
    uuid: str = field(init=False)
    confidence: float = 1.0

    def __post_init__(self) -> None:
        self.uuid = str(uuid4())


class DetectorSimulation(Detector):
    """This detector simulates object detection.

    It requires a camera provider in order to check visibility using the cameras' calibrations.
    Individual camera IDs can be added to a set of `blocked_cameras` to simulate occlusions during pytests.
    A list of `simulated_objects` can be filled to define what can be detected.
    An optional `noise` parameter controls the spatial accuracy in pixels.
    An optional `detection_delay` parameter simulates the time it takes to process an image.
    """

    def __init__(self,
                 camera_provider: CalibratableCameraProvider, *,
                 noise: float = 1.0,
                 detection_delay: float = 0.4,
                 name: str | None = None,
                 ) -> None:
        super().__init__(name=name)

        self.camera_provider = camera_provider
        self.noise = noise

        self.blocked_cameras: set[str] = set()
        self.simulated_objects: list[SimulatedObject] = []
        self.detection_delay = detection_delay

    async def detect(self,
                     image: Image,
                     *,
                     autoupload: Autoupload = Autoupload.FILTERED,
                     tags: list[str] | None = None,
                     source: str | None = None,
                     creation_date: datetime | str | None = None,
                     ) -> Detections | None:
        is_blocked = image.camera_id in self.blocked_cameras
        await rosys.sleep(self.detection_delay)
        image.set_detections(self.name, Detections())
        if not is_blocked:
            self.update_simulated_objects(image)
            self.detect_from_simulated_objects(image)
        self.NEW_DETECTIONS.emit(image)
        return image.get_detections(self.name)

    async def upload(self,
                     image: Image,
                     *,
                     tags: list[str] | None = None,
                     source: str | None = None,
                     creation_date: datetime | str | None = None,
                     ) -> None:
        self.log.info('Uploading %s', image.id)

    async def fetch_detector_info(self) -> DetectorInfo:
        return DetectorInfo(operation_mode='simulation',
                            version_control='pause',
                            state=None,
                            categories=[])

    async def fetch_model_version_info(self) -> ModelVersioningInfo:
        return ModelVersioningInfo(current_version='None',
                                   target_version='None',
                                   loop_version='None',
                                   local_versions=[],
                                   version_control='pause')

    async def set_model_version(self, version: Literal['follow_loop', 'pause'] | str) -> None:
        self.log.info('Setting model version to %s', version)

    def update_simulated_objects(self, image: Image) -> None:
        pass

    def detect_from_simulated_objects(self, image: Image) -> None:
        if image.camera_id not in self.camera_provider.cameras:
            return
        camera = self.camera_provider.cameras[image.camera_id]
        assert isinstance(camera, CalibratableCamera)
        assert camera.calibration is not None
        detections = image.get_detections(self.name)
        assert detections is not None
        world_extrinsics = camera.calibration.extrinsics.resolve()
        for obj in self.simulated_objects:
            viewing_direction = np.array(world_extrinsics.rotation.R)[:, 2]
            object_direction = np.array(obj.position.tuple) - world_extrinsics.translation
            if np.dot(viewing_direction, object_direction) < 0:
                continue
            image_point = camera.calibration.project_to_image(obj.position)
            if image_point is None:
                continue
            if not (0 <= image_point.x < image.size.width and 0 <= image_point.y < image.size.height):
                continue

            if obj.size is None:
                detections.points.append(PointDetection(
                    category_name=obj.category_name,
                    model_name='simulation',
                    confidence=obj.confidence,
                    x=image_point.x + self.noise * np.random.randn(),
                    y=image_point.y + self.noise * np.random.randn(),
                    uuid=obj.uuid,
                ))
            else:
                world_points = np.array([
                    [obj.position.x + dx, obj.position.y + dy, obj.position.z + dz]
                    for dx in [-obj.size[0] / 2, obj.size[0] / 2]
                    for dy in [-obj.size[1] / 2, obj.size[1] / 2]
                    for dz in [-obj.size[2] / 2, obj.size[2] / 2]
                ])
                image_points = camera.calibration.project_to_image(world_points)
                if np.any(np.isnan(image_points)):
                    continue
                detections.boxes.append(BoxDetection(
                    category_name=obj.category_name,
                    model_name='simulation',
                    confidence=obj.confidence,
                    x=image_points[:, 0].min() + self.noise * np.random.randn(),
                    y=image_points[:, 1].min() + self.noise * np.random.randn(),
                    width=image_points[:, 0].max() - image_points[:, 0].min() + self.noise * np.random.randn(),
                    height=image_points[:, 1].max() - image_points[:, 1].min() + self.noise * np.random.randn(),
                    uuid=obj.uuid,
                ))



from __future__ import annotations

import logging
import re
from collections.abc import Awaitable, Callable

import cv2
import numpy as np

from ... import rosys
from .usb_camera_scanner import device_nodes_from_uid

MJPG = cv2.VideoWriter.fourcc(*'MJPG')


def find_video_id(camera_uid: str) -> int | None:
    device_nodes = device_nodes_from_uid(camera_uid)
    video_ids = [
        int(node.strip().removeprefix('/dev/video'))
        for node in device_nodes
        if node.startswith('/dev/video')
    ]
    return min(video_ids) if video_ids else None


def to_bytes(image: np.ndarray) -> bytes:
    return image.tobytes()


class UsbDevice:

    def __init__(self, video_id: int, capture: cv2.VideoCapture, *,
                 on_new_image_data: Callable[[np.ndarray | bytes, float], Awaitable | None]) -> None:
        self._video_id: int = video_id
        self._capture: cv2.VideoCapture = capture
        self._on_new_image_data = on_new_image_data
        self._exposure_min: int = 0
        self._exposure_max: int = 0
        self._exposure_default: int = 0
        self._has_manual_exposure: bool = False
        self._video_formats: set[str] = set()
        self._image_is_jpg: bool = False

        self.set_video_format()

        self._capture_task = rosys.on_repeat(self._capture_image, interval=0.01)

    def __del__(self) -> None:
        self._capture.release()
        self._capture_task.stop()

    @property
    def video_formats(self) -> set[str]:
        return self._video_formats

    @property
    def image_is_jpg(self) -> bool:
        return self._image_is_jpg

    @staticmethod
    def from_uid(camera_id: str, on_new_image_data: Callable[[np.ndarray | bytes, float], Awaitable | None]) -> UsbDevice | None:
        video_id = find_video_id(camera_id)
        if video_id is None:
            logging.error('Could not find video device for camera %s', camera_id)
            return None

        capture = UsbDevice.create_capture(video_id)
        if capture is None:
            logging.error('Could not open video device %s', video_id)
            return None

        return UsbDevice(video_id=video_id, capture=capture, on_new_image_data=on_new_image_data)

    @staticmethod
    def create_capture(index: int) -> cv2.VideoCapture | None:
        capture = cv2.VideoCapture(index)
        if capture is None:
            return None
        capture.set(cv2.CAP_PROP_BUFFERSIZE, 1)
        if not capture.isOpened():
            capture.release()
            return None
        return capture

    async def _capture_image(self) -> None:
        if not self._capture.isOpened():
            return
        read_result = await rosys.run.io_bound(self._capture.read)
        if read_result is None:
            return
        capture_success, frame = read_result

        if capture_success:
            timestamp = rosys.time()
            if self._image_is_jpg:
                bytes_ = await rosys.run.io_bound(to_bytes, frame[0])
                if bytes_ is None:
                    return
                result = self._on_new_image_data(bytes_, timestamp)
            else:
                result = self._on_new_image_data(frame, timestamp)
            if isinstance(result, Awaitable):
                await result

    async def release_capture(self) -> None:
        await rosys.run.io_bound(self._capture.release)

    async def load_value_ranges(self) -> None:
        output = await self.run_v4l('--all')
        if output is None:
            return
        match = re.search(r'exposure_absolute.*: min=(\d*).*max=(\d*).*default=(\d*).*', output)
        if match is not None:
            self._has_manual_exposure = True
            self._exposure_min = int(match.group(1))
            self._exposure_max = int(match.group(2))
            self._exposure_default = int(match.group(3))
        else:
            self._has_manual_exposure = False
        output = await self.run_v4l('--list-formats')
        matches = re.finditer(r"$.*'(.*)'.*", output)
        for m in matches:
            self._video_formats.add(m.group(1))

    async def run_v4l(self, *args) -> str:
        cmd = ['v4l2-ctl', '-d', str(self._video_id)]
        cmd.extend(args)
        return await rosys.run.sh(cmd)

    def set_video_format(self) -> None:
        if 'MJPG' in self._video_formats:
            # NOTE enforcing motion jpeg for now
            if self._capture.get(cv2.CAP_PROP_FOURCC) != MJPG:
                self._capture.set(cv2.CAP_PROP_FOURCC, MJPG)
            # NOTE disable video decoding (see https://stackoverflow.com/questions/62664621/read-jpeg-frame-from-mjpeg-self-without-decoding-in-python-opencv/70869738?noredirect=1#comment110818859_62664621)
            if self._capture.get(cv2.CAP_PROP_CONVERT_RGB) != 0:
                self._capture.set(cv2.CAP_PROP_CONVERT_RGB, 0)
            # NOTE make sure there is no lag (see https://stackoverflow.com/a/30032945/364388)
            if self._capture.get(cv2.CAP_PROP_BUFFERSIZE) != 1:
                self._capture.set(cv2.CAP_PROP_BUFFERSIZE, 1)
            self._image_is_jpg = True
        else:
            self._image_is_jpg = False

    def set_auto_exposure(self, auto: bool) -> None:
        if self._has_manual_exposure:
            if auto:
                self._capture.set(cv2.CAP_PROP_AUTO_EXPOSURE, 3)
            else:
                self._capture.set(cv2.CAP_PROP_AUTO_EXPOSURE, 1)

    def set_exposure(self, value: float) -> None:
        if self._has_manual_exposure:
            is_auto_exposure = self.get_auto_exposure()
            if not is_auto_exposure:
                exposure = self._capture.get(cv2.CAP_PROP_EXPOSURE) / self._exposure_max
                if value != exposure:
                    self._capture.set(cv2.CAP_PROP_EXPOSURE, int(value * self._exposure_max))

    def get_auto_exposure(self) -> bool | None:
        return self._capture.get(cv2.CAP_PROP_AUTO_EXPOSURE) == 3

    def get_exposure(self) -> float | None:
        if not self._has_manual_exposure:
            return None
        return self._capture.get(cv2.CAP_PROP_EXPOSURE) / self._exposure_max

    def set_width(self, width: int) -> None:
        self._capture.set(cv2.CAP_PROP_FRAME_WIDTH, width)

    def get_width(self) -> int:
        return int(self._capture.get(cv2.CAP_PROP_FRAME_WIDTH))

    def set_height(self, height: int) -> None:
        self._capture.set(cv2.CAP_PROP_FRAME_HEIGHT, height)

    def get_height(self) -> int:
        return int(self._capture.get(cv2.CAP_PROP_FRAME_HEIGHT))

    def set_fps(self, fps: int) -> None:
        self._capture.set(cv2.CAP_PROP_FPS, fps)

    def get_fps(self) -> int:
        return int(self._capture.get(cv2.CAP_PROP_FPS))
